@@ -20,7 +20,7 @@ cmds = {
 
 def get_sdr_phy_init_sequence(phy_settings, timing_settings):
     cl = phy_settings.cl
-    bl = 1
+    bl = phy_settings.nphases
     mr = log2_int(bl) + (cl << 4)
     reset_dll = 1 << 8
 
@@ -151,12 +151,13 @@ def get_ddr3_phy_init_sequence(phy_settings, timing_settings):
         mr0 |= wr_to_mr0[wr] << 9
         return mr0
 
-    def format_mr1(ron, rtt_nom):
+    def format_mr1(ron, rtt_nom, tdqs):
         mr1 = ((ron >> 0) & 1) << 1
         mr1 |= ((ron >> 1) & 1) << 5
         mr1 |= ((rtt_nom >> 0) & 1) << 2
         mr1 |= ((rtt_nom >> 1) & 1) << 6
         mr1 |= ((rtt_nom >> 2) & 1) << 9
+        mr1 |= (tdqs & 1) << 11
         return mr1
 
     def format_mr2(cwl, rtt_wr):
@@ -188,6 +189,7 @@ def get_ddr3_phy_init_sequence(phy_settings, timing_settings):
     rtt_nom = "60ohm"
     rtt_wr  = "60ohm"
     ron     = "34ohm"
+    tdqs    = 0
 
     # override electrical settings if specified
     if hasattr(phy_settings, "rtt_nom"):
@@ -196,10 +198,12 @@ def get_ddr3_phy_init_sequence(phy_settings, timing_settings):
         rtt_wr = phy_settings.rtt_wr
     if hasattr(phy_settings, "ron"):
         ron = phy_settings.ron
+    if getattr(phy_settings, "tdqs", False):
+        tdqs = 1
 
     wr  = max(timing_settings.tWTR*phy_settings.nphases, 5) # >= ceiling(tWR/tCK)
     mr0 = format_mr0(bl, cl, wr, 1)
-    mr1 = format_mr1(z_to_ron[ron], z_to_rtt_nom[rtt_nom])
+    mr1 = format_mr1(z_to_ron[ron], z_to_rtt_nom[rtt_nom], tdqs)
     mr2 = format_mr2(cwl, z_to_rtt_wr[rtt_wr])
     mr3 = 0
 
@@ -274,13 +278,14 @@ def get_ddr4_phy_init_sequence(phy_settings, timing_settings):
         mr0 |= (wr_to_mr0[wr] >> 3) << 13
         return mr0
 
-    def format_mr1(dll_enable, ron, rtt_nom):
+    def format_mr1(dll_enable, ron, rtt_nom, tdqs):
         mr1 = dll_enable
         mr1 |= ((ron >> 0) & 0b1) << 1
         mr1 |= ((ron >> 1) & 0b1) << 2
         mr1 |= ((rtt_nom >> 0) & 0b1) << 8
         mr1 |= ((rtt_nom >> 1) & 0b1) << 9
         mr1 |= ((rtt_nom >> 2) & 0b1) << 10
+        mr1 |= (tdqs & 0b1) << 11
         return mr1
 
     def format_mr2(cwl, rtt_wr):
@@ -346,6 +351,7 @@ def get_ddr4_phy_init_sequence(phy_settings, timing_settings):
     rtt_nom = "40ohm"
     rtt_wr  = "120ohm"
     ron     = "34ohm"
+    tdqs    = 0
 
     # override electrical settings if specified
     if hasattr(phy_settings, "rtt_nom"):
@@ -354,19 +360,74 @@ def get_ddr4_phy_init_sequence(phy_settings, timing_settings):
         rtt_wr = phy_settings.rtt_wr
     if hasattr(phy_settings, "ron"):
         ron = phy_settings.ron
+    if getattr(phy_settings, "tdqs", False):
+        tdqs = 1
 
     wr  = max(timing_settings.tWTR*phy_settings.nphases, 10) # >= ceiling(tWR/tCK)
     mr0 = format_mr0(bl, cl, wr, 1)
-    mr1 = format_mr1(1, z_to_ron[ron], z_to_rtt_nom[rtt_nom])
+    mr1 = format_mr1(1, z_to_ron[ron], z_to_rtt_nom[rtt_nom], tdqs)
     mr2 = format_mr2(cwl, z_to_rtt_wr[rtt_wr])
     mr3 = format_mr3(timing_settings.fine_refresh_mode)
     mr4 = 0
     mr5 = 0
     mr6 = format_mr6(4) # FIXME: tCCD
 
+    rdimm_init = []
+    if phy_settings.is_rdimm:
+        def get_coarse_speed(tck, pll_bypass):
+            # JESD82-31A page 78
+            f_to_coarse_speed = {
+                1600e6: 0,
+                1866e6: 1,
+                2133e6: 2,
+                2400e6: 3,
+                2666e6: 4,
+                2933e6: 5,
+                3200e6: 6,
+            }
+            if pll_bypass:
+                return 7
+            else:
+                for f, speed in f_to_coarse_speed.items():
+                        if tck >= 2/f:
+                            return speed
+                raise ValueError
+        def get_fine_speed(tck):
+            # JESD82-31A page 83
+            freq = 2/tck
+            fine_speed = (freq - 1240e6) // 20e6
+            fine_speed = max(fine_speed, 0)
+            fine_speed = min(fine_speed, 0b1100001)
+            return fine_speed
+
+        coarse_speed = get_coarse_speed(phy_settings.tck, phy_settings.rcd_pll_bypass)
+        fine_speed = get_fine_speed(phy_settings.tck)
+
+        rcd_reset = 0x060 | 0x0                          # F0RC06: command space control; 0: reset RCD
+
+        f0rc0f = 0x0F0 | 0x4                             # F0RC05: 0 nCK latency adder
+
+        f0rc03 = 0x030 | phy_settings.rcd_ca_cs_drive    # F0RC03: CA/CS drive strength
+        f0rc04 = 0x040 | phy_settings.rcd_odt_cke_drive  # F0RC04: ODT/CKE drive strength
+        f0rc05 = 0x050 | phy_settings.rcd_clk_drive      # F0RC04: ODT/CKE drive strength
+
+        f0rc0a = 0x0A0 | coarse_speed                    # F0RC0A: coarse speed selection and PLL bypass
+        f0rc3x = 0x300 | fine_speed                      # F0RC3x: fine speed selection
+
+        rdimm_init = [
+            ("Reset RCD", rcd_reset, 7, cmds["MODE_REGISTER"], 50000),
+            ("Load RCD F0RC0F", f0rc0f, 7, cmds["MODE_REGISTER"], 100),
+            ("Load RCD F0RC03", f0rc03, 7, cmds["MODE_REGISTER"], 100),
+            ("Load RCD F0RC04", f0rc04, 7, cmds["MODE_REGISTER"], 100),
+            ("Load RCD F0RC05", f0rc05, 7, cmds["MODE_REGISTER"], 100),
+            ("Load RCD F0RC0A", f0rc0a, 7, cmds["MODE_REGISTER"], 100),
+            ("Load RCD F0RC3X", f0rc3x, 7, cmds["MODE_REGISTER"], 100),
+        ]
+
     init_sequence = [
         ("Release reset", 0x0000, 0, cmds["UNRESET"], 50000),
         ("Bring CKE high", 0x0000, 0, cmds["CKE"], 10000),
+    ] + rdimm_init + [
         ("Load Mode Register 3", mr3, 3, cmds["MODE_REGISTER"], 0),
         ("Load Mode Register 6", mr6, 6, cmds["MODE_REGISTER"], 0),
         ("Load Mode Register 5", mr5, 5, cmds["MODE_REGISTER"], 0),
@@ -395,13 +456,31 @@ def get_sdram_phy_init_sequence(phy_settings, timing_settings):
 
 def get_sdram_phy_c_header(phy_settings, timing_settings):
     r = "#ifndef __GENERATED_SDRAM_PHY_H\n#define __GENERATED_SDRAM_PHY_H\n"
-    r += "#include <hw/common.h>\n#include <generated/csr.h>\n#include <hw/flags.h>\n\n"
+    r += "#include <hw/common.h>\n"
+    r += "#include <generated/csr.h>\n"
+    r += "\n"
+
+    r += "#define DFII_CONTROL_SEL        0x01\n"
+    r += "#define DFII_CONTROL_CKE        0x02\n"
+    r += "#define DFII_CONTROL_ODT        0x04\n"
+    r += "#define DFII_CONTROL_RESET_N    0x08\n"
+    r += "\n"
+
+    r += "#define DFII_COMMAND_CS         0x01\n"
+    r += "#define DFII_COMMAND_WE         0x02\n"
+    r += "#define DFII_COMMAND_CAS        0x04\n"
+    r += "#define DFII_COMMAND_RAS        0x08\n"
+    r += "#define DFII_COMMAND_WRDATA     0x10\n"
+    r += "#define DFII_COMMAND_RDDATA     0x20\n"
+    r += "\n"
 
     phytype = phy_settings.phytype.upper()
     nphases = phy_settings.nphases
 
     # Define PHY type and number of phases
     r += "#define SDRAM_PHY_"+phytype+"\n"
+    r += "#define SDRAM_PHY_XDR "+str(1 if phy_settings.memtype == "SDR" else 2) + "\n"
+    r += "#define SDRAM_PHY_DATABITS "+str(phy_settings.databits) + "\n"
     r += "#define SDRAM_PHY_PHASES "+str(nphases)+"\n"
 
     # Define Read/Write Leveling capability
@@ -416,16 +495,20 @@ def get_sdram_phy_c_header(phy_settings, timing_settings):
     if phytype in ["USDDRPHY", "USPDDRPHY"]:
         r += "#define SDRAM_PHY_MODULES DFII_PIX_DATA_BYTES/2\n"
         r += "#define SDRAM_PHY_DELAYS 512\n"
-        r += "#define SDRAM_PHY_BITSLIPS 8\n"
+        r += "#define SDRAM_PHY_BITSLIPS 16\n"
     elif phytype in ["A7DDRPHY", "K7DDRPHY", "V7DDRPHY"]:
         r += "#define SDRAM_PHY_MODULES DFII_PIX_DATA_BYTES/2\n"
         r += "#define SDRAM_PHY_DELAYS 32\n"
-        r += "#define SDRAM_PHY_BITSLIPS 8\n"
+        r += "#define SDRAM_PHY_BITSLIPS 16\n"
     elif phytype in ["ECP5DDRPHY"]:
         r += "#define SDRAM_PHY_MODULES DFII_PIX_DATA_BYTES/4\n"
         r += "#define SDRAM_PHY_DELAYS 8\n"
         r += "#define SDRAM_PHY_BITSLIPS 4\n"
 
+    if phy_settings.is_rdimm:
+        assert phy_settings.memtype == "DDR4"
+        r += "#define SDRAM_PHY_DDR4_RDIMM\n"
+
     r += "\n"
 
     r += "static void cdelay(int i);\n"
@@ -482,16 +565,32 @@ const unsigned long sdram_dfii_pix_rddata_addr[SDRAM_PHY_PHASES] = {{
 
     r += "static void init_sequence(void)\n{\n"
     for comment, a, ba, cmd, delay in init_sequence:
-        r += "\t/* {0} */\n".format(comment)
-        r += "\tsdram_dfii_pi0_address_write({0:#x});\n".format(a)
-        r += "\tsdram_dfii_pi0_baddress_write({0:d});\n".format(ba)
-        if cmd[:12] == "DFII_CONTROL":
-            r += "\tsdram_dfii_control_write({0});\n".format(cmd)
-        else:
-            r += "\tcommand_p0({0});\n".format(cmd)
-        if delay:
-            r += "\tcdelay({0:d});\n".format(delay)
-        r += "\n"
+        invert_masks = [(0, 0), ]
+        if phy_settings.is_rdimm:
+            assert phy_settings.memtype == "DDR4"
+            # JESD82-31A page 38
+            #
+            # B-side chips have certain usually-inconsequential address and BA
+            # bits inverted by the RCD to reduce SSO current. For mode register
+            # writes, however, we must compensate for this. BG[1] also directs
+            # writes either to the A side (BG[1]=0) or B side (BG[1]=1)
+            #
+            # The 'ba != 7' is because we don't do this to writes to the RCD
+            # itself.
+            if ba != 7:
+                invert_masks.append((0b10101111111000, 0b1111))
+
+        for a_inv, ba_inv in invert_masks:
+            r += "\t/* {0} */\n".format(comment)
+            r += "\tsdram_dfii_pi0_address_write({0:#x});\n".format(a ^ a_inv)
+            r += "\tsdram_dfii_pi0_baddress_write({0:d});\n".format(ba ^ ba_inv)
+            if cmd[:12] == "DFII_CONTROL":
+                r += "\tsdram_dfii_control_write({0});\n".format(cmd)
+            else:
+                r += "\tcommand_p0({0});\n".format(cmd)
+            if delay:
+                r += "\tcdelay({0:d});\n".format(delay)
+            r += "\n"
     r += "}\n"
 
     r += "#endif\n"
