@@ -48,46 +48,49 @@ def _twos_complement(value, nbits):
 def _word(msb, lsb):
     return (msb << 8) | lsb
 
+# most signifficant (upper) / least signifficant (lower) nibble
+def _msn(byte):
+    return _read_field(byte, nbits=4, shift=4)
+
+def _lsn(byte):
+    return _read_field(byte, nbits=4, shift=0)
+
 
 class DDR3SPDData:
     memtype = "DDR3"
+    _speedgrades = [800, 1066, 1333, 1600, 1866, 2133]
 
     def __init__(self, spd_data):
-        # Geometry ---------------------------------------------------------------------------------
+        self.get_geometry(spd_data)
+        self.init_timebase(spd_data)
+        self.get_timings(spd_data)
+
+    def get_geometry(self, data):
         bankbits = {
             0b000: 3,
             0b001: 4,
             0b010: 5,
             0b011: 6,
-        }[_read_field(spd_data[4], nbits=3, shift=4)]
+        }[_read_field(data[4], nbits=3, shift=4)]
         rowbits = {
             0b000: 12,
             0b001: 13,
             0b010: 14,
             0b011: 15,
             0b100: 16,
-        }[_read_field(spd_data[5], nbits=3, shift=3)]
+        }[_read_field(data[5], nbits=3, shift=3)]
         colbits = {
             0b000:  9,
             0b001: 10,
             0b010: 11,
             0b011: 12,
-        }[_read_field(spd_data[5], nbits=3, shift=0)]
+        }[_read_field(data[5], nbits=3, shift=0)]
 
         self.nbanks = 2**bankbits
         self.nrows = 2**rowbits
         self.ncols = 2**colbits
 
-        # Timings ----------------------------------------------------------------------------------
-        self.init_timebase(spd_data)
-
-        # most signifficant (upper) / least signifficant (lower) nibble
-        def msn(byte):
-            return _read_field(byte, nbits=4, shift=4)
-
-        def lsn(byte):
-            return _read_field(byte, nbits=4, shift=0)
-
+    def get_timings(self, spd_data):
         b = spd_data
         tck_min  = self.txx_ns(mtb=b[12], ftb=b[34])
         taa_min  = self.txx_ns(mtb=b[16], ftb=b[35])
@@ -95,12 +98,12 @@ class DDR3SPDData:
         trcd_min = self.txx_ns(mtb=b[18], ftb=b[36])
         trrd_min = self.txx_ns(mtb=b[19])
         trp_min  = self.txx_ns(mtb=b[20], ftb=b[37])
-        tras_min = self.txx_ns(mtb=_word(lsn(b[21]), b[22]))
-        trc_min  = self.txx_ns(mtb=_word(msn(b[21]), b[23]), ftb=b[38])
+        tras_min = self.txx_ns(mtb=_word(_lsn(b[21]), b[22]))
+        trc_min  = self.txx_ns(mtb=_word(_msn(b[21]), b[23]), ftb=b[38])
         trfc_min = self.txx_ns(mtb=_word(b[25], b[24]))
         twtr_min = self.txx_ns(mtb=b[26])
         trtp_min = self.txx_ns(mtb=b[27])
-        tfaw_min = self.txx_ns(mtb=_word(lsn(b[28]), b[29]))
+        tfaw_min = self.txx_ns(mtb=_word(_lsn(b[28]), b[29]))
 
         technology_timings = _TechnologyTimings(
             tREFI = 64e6/8192,      # 64ms/8192ops
@@ -142,23 +145,150 @@ class DDR3SPDData:
         ftb = _twos_complement(ftb, 8)
         return mtb * self.medium_timebase_ns + ftb * self.fine_timebase_ns
 
-    @staticmethod
-    def speedgrade_freq(tck_ns):
+    @classmethod
+    def speedgrade_freq(cls, tck_ns):
         # Calculate rounded speedgrade frequency from tck_min
         freq_mhz = (1 / (tck_ns * 1e-9)) / 1e6
         freq_mhz *= 2  # clock rate -> transfer rate (DDR)
-        speedgrades = [800, 1066, 1333, 1600, 1866, 2133]
-        for f in speedgrades:
+        for f in cls._speedgrades:
             # Due to limited tck accuracy of 1ps, calculations may yield higher
             # frequency than in reality (e.g. for DDR3-1866: tck=1.071 ns ->
             # -> f=1867.4 MHz, while real is f=1866.6(6) MHz).
             max_error = 2
             if abs(freq_mhz - f) < max_error:
                 return f
-        raise ValueError("Transfer rate = {:.2f} does not correspond to any DDR3 speedgrade"
+        raise ValueError("Transfer rate = {:.2f} does not correspond to any speedgrade"
                          .format(freq_mhz))
 
 
+class DDR4SPDData(DDR3SPDData):
+    memtype = "DDR4"
+    _speedgrades = [1600, 1866, 2133, 2400, 2666, 2933, 3200]
+
+    def get_geometry(self, data):
+        bankgroupbits = {
+            0b00: 0,
+            0b01: 1,
+            0b10: 2,
+        }[_read_field(data[4], nbits=2, shift=6)]
+        bankbits = {
+            0b00: 2,
+            0b01: 3,
+        }[_read_field(data[4], nbits=2, shift=4)]
+        rowbits = {
+            0b000: 12,
+            0b001: 13,
+            0b010: 14,
+            0b011: 15,
+            0b100: 16,
+            0b101: 17,
+            0b110: 18,
+        }[_read_field(data[5], nbits=3, shift=3)]
+        colbits = {
+            0b000:  9,
+            0b001: 10,
+            0b010: 11,
+            0b011: 12,
+        }[_read_field(data[5], nbits=3, shift=0)]
+
+        self.ngroups = 2**bankgroupbits
+        self.ngroupbanks = 2**bankbits
+        self.nbanks = self.ngroups * self.ngroupbanks
+        self.nrows = 2**rowbits
+        self.ncols = 2**colbits
+
+    def init_timebase(self, data):
+        # there is only one possible value for DDR4
+        self.medium_timebase_ns = {
+            0b00: 125e-3,
+        }[_read_field(data[17], nbits=2, shift=2)]
+        self.fine_timebase_ns = {
+            0b00: 1e-3,
+        }[_read_field(data[17], nbits=2, shift=0)]
+
+    def get_timings(self, data):
+        b = data
+
+        self.trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
+
+        tckavg_min  = self.txx_ns(mtb=b[18], ftb=b[125])
+        tckavg_max  = self.txx_ns(mtb=b[19], ftb=b[124])
+        taa_min  = self.txx_ns(mtb=b[24], ftb=b[123])
+        trcd_min = self.txx_ns(mtb=b[25], ftb=b[122])
+        trp_min  = self.txx_ns(mtb=b[26], ftb=b[121])
+        tras_min = self.txx_ns(mtb=_word(_lsn(b[27]), b[28]))
+        trc_min  = self.txx_ns(mtb=_word(_msn(b[27]), b[29]), ftb=b[120])
+        self.trfc = {
+            "1x": (None, self.txx_ns(mtb=_word(b[31], b[30]))),
+            "2x": (None, self.txx_ns(mtb=_word(b[33], b[32]))),
+            "4x": (None, self.txx_ns(mtb=_word(b[35], b[34]))),
+        }
+        tfaw_min = self.txx_ns(mtb=_word(_lsn(b[36]), b[37]))
+        trrd_s_min = self.txx_ns(mtb=b[38], ftb=b[119])
+        trrd_l_min = self.txx_ns(mtb=b[39], ftb=b[118])
+        tccd_l_min = self.txx_ns(mtb=b[40], ftb=b[117])  # min 6 cycles?
+        twr_min  = self.txx_ns(mtb=_word(_lsn(b[41]), b[42]))
+        twtr_s_min = self.txx_ns(mtb=_word(_lsn(b[43]), b[44]))
+        twtr_l_min = self.txx_ns(mtb=_word(_msn(b[43]), b[45]))
+
+        # minimum tFAW in clock cycles depends on page size
+        sdram_device_width = {
+            0b000:  4,
+            0b001:  8,
+            0b010: 16,
+            0b011: 32,
+        }[_read_field(b[12], nbits=3, shift=0)]
+        page_size_bytes = self.ncols * sdram_device_width / 8
+        tfaw_min_ck = {
+            512:  16,
+            1024: 20,
+            2048: 28,
+        }[page_size_bytes]
+
+        technology_timings = _TechnologyTimings(
+            tREFI = self.trefi,
+            tWTR  = (4, twtr_l_min),
+            tCCD  = (4, tccd_l_min),
+            tRRD  = (4, trrd_l_min),
+            tZQCS = (128, 80),
+        )
+        speedgrade_timings = _SpeedgradeTimings(
+            tRP  = trp_min,
+            tRCD = trcd_min,
+            tWR  = twr_min,
+            tRFC = self.trfc,
+            tFAW = (tfaw_min_ck, tfaw_min),
+            tRAS = tras_min,
+        )
+
+        self.speedgrade = str(self.speedgrade_freq(tckavg_min))
+        self.technology_timings = technology_timings
+        self.speedgrade_timings = {
+            self.speedgrade: speedgrade_timings,
+            "default": speedgrade_timings,
+        }
+
+def parse_spd_hexdump(filename):
+    """Parse data dumped using the `spdread` command in LiteX BIOS
+
+    This will read files in format:
+        Memory dump:
+        0x00000000  00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f  ................
+        0x00000010  10 11 12 13 14 15 16 17 18 19 1a 1b 1c 1d 1e 1f  ................
+    """
+    data = []
+    last_addr = -1
+    with open(filename) as f:
+        for line in f:
+            if line.startswith("0x"):
+                tokens = line.strip().split()
+                addr = int(tokens[0], 16)
+                assert addr > last_addr
+                values = [int(v, 16) for v in tokens[1:17]]
+                data.extend(values)
+                last_addr = addr
+    return data
+
 # SDRAMModule --------------------------------------------------------------------------------------
 
 class SDRAMModule:
@@ -172,6 +302,7 @@ class SDRAMModule:
     SDRAM modules with the same geometry exist can have
     various speedgrades.
     """
+    registered = False
     def __init__(self, clk_freq, rate, speedgrade=None, fine_refresh_mode=None):
         self.clk_freq      = clk_freq
         self.rate          = rate
@@ -254,6 +385,7 @@ class SDRAMModule:
         # set parameters from SPD data based on memory type
         spd_cls = {
             0x0b: DDR3SPDData,
+            0x0c: DDR4SPDData,
         }[spd_data[2]]
         spd = spd_cls(spd_data)
 
@@ -265,6 +397,8 @@ class SDRAMModule:
             ncols = spd.ncols
             technology_timings = spd.technology_timings
             speedgrade_timings = spd.speedgrade_timings
+            # Save data for runtime verification
+            _spd_data = spd_data
 
         nphases = {
             "SDR":   1,
@@ -276,13 +410,19 @@ class SDRAMModule:
         }[spd.memtype]
         rate = "1:{}".format(nphases)
 
-        return _SDRAMModule(clk_freq, rate=rate, speedgrade=spd.speedgrade,
-                            fine_refresh_mode=fine_refresh_mode)
+        return _SDRAMModule(clk_freq,
+            rate              = rate,
+            speedgrade        = spd.speedgrade,
+            fine_refresh_mode = fine_refresh_mode)
+
+class SDRAMRegisteredModule(SDRAMModule): registered = True
 
 # SDR ----------------------------------------------------------------------------------------------
 
-class IS42S16160(SDRAMModule):
-    memtype = "SDR"
+class SDRModule(SDRAMModule):                     memtype = "SDR"
+class SDRRegisteredModule(SDRAMRegisteredModule): memtype = "SDR"
+
+class IS42S16160(SDRModule):
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -291,9 +431,7 @@ class IS42S16160(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=20, tRFC=(None, 70), tFAW=None, tRAS=None)}
 
-
-class IS42S16320(SDRAMModule):
-    memtype = "SDR"
+class IS42S16320(SDRModule):
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -302,9 +440,7 @@ class IS42S16320(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=20, tRFC=(None, 70), tFAW=None, tRAS=None)}
 
-
-class MT48LC4M16(SDRAMModule):
-    memtype = "SDR"
+class MT48LC4M16(SDRModule):
     # geometry
     nbanks = 4
     nrows  = 4096
@@ -313,9 +449,7 @@ class MT48LC4M16(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=14, tRFC=(None, 66), tFAW=None, tRAS=None)}
 
-
-class MT48LC16M16(SDRAMModule):
-    memtype = "SDR"
+class MT48LC16M16(SDRModule):
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -324,9 +458,7 @@ class MT48LC16M16(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 15))
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=15, tRFC=(None, 66), tFAW=None, tRAS=44)}
 
-
-class AS4C16M16(SDRAMModule):
-    memtype = "SDR"
+class AS4C16M16(SDRModule):
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -335,9 +467,7 @@ class AS4C16M16(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=18, tRCD=18, tWR=12, tRFC=(None, 60), tFAW=None, tRAS=None)}
 
-
-class AS4C32M16(SDRAMModule):
-    memtype = "SDR"
+class AS4C32M16(SDRModule):
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -346,8 +476,7 @@ class AS4C32M16(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=18, tRCD=18, tWR=12, tRFC=(None, 60), tFAW=None, tRAS=None)}
 
-class AS4C32M8(SDRAMModule):
-    memtype = "SDR"
+class AS4C32M8(SDRModule):
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -356,8 +485,7 @@ class AS4C32M8(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 15))
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=15, tRFC=(None, 66), tFAW=None, tRAS=44)}
 
-class M12L64322A(SDRAMModule):
-    memtype = "SDR"
+class M12L64322A(SDRModule):
     # geometry
     nbanks = 4
     nrows  = 2048
@@ -366,8 +494,7 @@ class M12L64322A(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/4096, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 10))
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 55), tFAW=None, tRAS=40)}
 
-class M12L16161A(SDRAMModule):
-    memtype = "SDR"
+class M12L16161A(SDRModule):
     # geometry
     nbanks = 2
     nrows  = 2048
@@ -378,6 +505,9 @@ class M12L16161A(SDRAMModule):
 
 # DDR ----------------------------------------------------------------------------------------------
 
+class DDRModule(SDRAMModule):                     memtype = "DDR"
+class DDRRegisteredModule(SDRAMRegisteredModule): memtype = "DDR"
+
 class MT46V32M16(SDRAMModule):
     memtype = "DDR"
     # geometry
@@ -388,10 +518,12 @@ class MT46V32M16(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 70), tFAW=None, tRAS=None)}
 
+# LPDDR --------------------------------------------------------------------------------------------
 
-# LPDDR
-class MT46H32M16(SDRAMModule):
-    memtype = "LPDDR"
+class LPDDRModule(SDRAMModule):                     memtype = "LPDDR"
+class LPDDRRegisteredModule(SDRAMRegisteredModule): memtype = "LPDDR"
+
+class MT46H32M16(LPDDRModule):
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -400,9 +532,7 @@ class MT46H32M16(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 72), tFAW=None, tRAS=None)}
 
-
-class MT46H32M32(SDRAMModule):
-    memtype = "LPDDR"
+class MT46H32M32(LPDDRModule):
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -411,9 +541,12 @@ class MT46H32M32(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 72), tFAW=None, tRAS=None)}
 
-
 # DDR2 ---------------------------------------------------------------------------------------------
-class MT47H128M8(SDRAMModule):
+
+class DDR2Module(SDRAMModule):                     memtype = "DDR2"
+class DDR2RegisteredModule(SDRAMRegisteredModule): memtype = "DDR2"
+
+class MT47H128M8(DDR2Module):
     memtype = "DDR2"
     # geometry
     nbanks = 8
@@ -423,8 +556,7 @@ class MT47H128M8(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(None, 7.5), tCCD=(2, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}
 
-
-class MT47H32M16(SDRAMModule):
+class MT47H32M16(DDR2Module):
     memtype = "DDR2"
     # geometry
     nbanks = 4
@@ -434,8 +566,7 @@ class MT47H32M16(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(None, 7.5), tCCD=(2, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}
 
-
-class MT47H64M16(SDRAMModule):
+class MT47H64M16(DDR2Module):
     memtype = "DDR2"
     # geometry
     nbanks = 8
@@ -445,8 +576,7 @@ class MT47H64M16(SDRAMModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(None, 7.5), tCCD=(2, None), tRRD=None)
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}
 
-
-class P3R1GE4JGF(SDRAMModule):
+class P3R1GE4JGF(DDR2Module):
     memtype = "DDR2"
     # geometry
     nbanks = 8
@@ -458,7 +588,10 @@ class P3R1GE4JGF(SDRAMModule):
 
 # DDR3 (Chips) -------------------------------------------------------------------------------------
 
-class MT41K64M16(SDRAMModule):
+class DDR3Module(SDRAMModule):                     memtype = "DDR3"
+class DDR3RegisteredModule(SDRAMRegisteredModule): memtype = "DDR3"
+
+class MT41K64M16(DDR3Module):
     memtype = "DDR3"
     # geometry
     nbanks = 8
@@ -474,8 +607,7 @@ class MT41K64M16(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
-
-class MT41J128M16(SDRAMModule):
+class MT41J128M16(DDR3Module):
     memtype = "DDR3"
     # geometry
     nbanks = 8
@@ -491,13 +623,9 @@ class MT41J128M16(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
+class MT41K128M16(MT41J128M16): pass
 
-class MT41K128M16(MT41J128M16):
-    pass
-
-
-class MT41J256M16(SDRAMModule):
-    memtype = "DDR3"
+class MT41J256M16(DDR3Module):
     # geometry
     nbanks = 8
     nrows  = 32768
@@ -512,13 +640,23 @@ class MT41J256M16(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
+class MT41K256M16(MT41J256M16): pass
 
-class MT41K256M16(MT41J256M16):
-    pass
+class MT41J512M16(DDR3Module):
+    # geometry
+    nbanks = 8
+    nrows  = 65536
+    ncols  = 1024
+    # timings
+    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 10), tZQCS=(64, 80))
+    speedgrade_timings = {
+        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=13.75, tRFC=(280, None), tFAW=(None, 40), tRAS=39),
+    }
+    speedgrade_timings["default"] = speedgrade_timings["1600"]
 
+class MT41K512M16(MT41J512M16): pass
 
-class K4B1G0446F(SDRAMModule):
-    memtype = "DDR3"
+class K4B1G0446F(DDR3Module):
     # geometry
     nbanks = 8
     nrows  = 16384
@@ -533,9 +671,7 @@ class K4B1G0446F(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
-
-class K4B2G1646F(SDRAMModule):
-    memtype = "DDR3"
+class K4B2G1646F(DDR3Module):
     # geometry
     nbanks = 8
     nrows  = 16384
@@ -550,9 +686,7 @@ class K4B2G1646F(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
-
-class H5TC4G63CFR(SDRAMModule):
-    memtype = "DDR3"
+class H5TC4G63CFR(DDR3Module):
     # geometry
     nbanks = 8
     nrows  = 16384
@@ -564,9 +698,7 @@ class H5TC4G63CFR(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["800"]
 
-
-class IS43TR16128B(SDRAMModule):
-    memtype = "DDR3"
+class IS43TR16128B(DDR3Module):
     # geometry
     nbanks = 8
     nrows  = 16384
@@ -581,9 +713,8 @@ class IS43TR16128B(SDRAMModule):
 
 # DDR3 (SO-DIMM) -----------------------------------------------------------------------------------
 
-class MT8JTF12864(SDRAMModule):
+class MT8JTF12864(DDR3Module):
     # base chip: MT41J128M8
-    memtype = "DDR3"
     # geometry
     nbanks = 8
     nrows  = 16384
@@ -596,10 +727,8 @@ class MT8JTF12864(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["1333"]
 
-
-class MT8KTF51264(SDRAMModule):
+class MT8KTF51264(DDR3Module):
     # base chip: MT41K512M8
-    memtype = "DDR3"
     # geometry
     nbanks = 8
     nrows  = 65536
@@ -615,10 +744,8 @@ class MT8KTF51264(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["1866"]
 
-
-class MT18KSF1G72HZ(SDRAMModule):
+class MT18KSF1G72HZ(DDR3Module):
     # base chip: MT41K512M8
-    memtype = "DDR3"
     # geometry
     nbanks = 8
     nrows  = 65536
@@ -632,9 +759,7 @@ class MT18KSF1G72HZ(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
-
-class AS4C256M16D3A(SDRAMModule):
-    memtype = "DDR3"
+class AS4C256M16D3A(DDR3Module):
     # geometry
     nbanks = 8
     nrows  = 32768
@@ -646,10 +771,8 @@ class AS4C256M16D3A(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
-
-class MT16KTF1G64HZ(SDRAMModule):
+class MT16KTF1G64HZ(DDR3Module):
     # base chip: MT41K512M8
-    memtype = "DDR3"
     # geometry
     nbanks = 8
     nrows  = 65536
@@ -667,8 +790,11 @@ class MT16KTF1G64HZ(SDRAMModule):
 
 
 # DDR4 (Chips) -------------------------------------------------------------------------------------
-class EDY4016A(SDRAMModule):
-    memtype = "DDR4"
+
+class DDR4Module(SDRAMModule):                     memtype = "DDR4"
+class DDR4RegisteredModule(SDRAMRegisteredModule): memtype = "DDR4"
+
+class EDY4016A(DDR4Module):
     # geometry
     ngroupbanks = 4
     ngroups     = 2
@@ -684,9 +810,7 @@ class EDY4016A(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["2400"]
 
-
-class MT40A1G8(SDRAMModule):
-    memtype = "DDR4"
+class MT40A1G8(DDR4Module):
     # geometry
     ngroupbanks = 4
     ngroups     = 4
@@ -703,9 +827,7 @@ class MT40A1G8(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["2400"]
 
-
-class MT40A256M16(SDRAMModule):
-    memtype = "DDR4"
+class MT40A256M16(DDR4Module):
     # geometry
     ngroupbanks = 4
     ngroups     = 2
@@ -721,9 +843,7 @@ class MT40A256M16(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["2400"]
 
-
-class MT40A512M8(SDRAMModule):
-    memtype = "DDR4"
+class MT40A512M8(DDR4Module):
     # geometry
     ngroupbanks = 4
     ngroups     = 4
@@ -740,9 +860,7 @@ class MT40A512M8(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["2400"]
 
-
-class MT40A512M16(SDRAMModule):
-    memtype = "DDR4"
+class MT40A512M16(DDR4Module):
     # geometry
     ngroupbanks = 4
     ngroups     = 2
@@ -759,8 +877,8 @@ class MT40A512M16(SDRAMModule):
     speedgrade_timings["default"] = speedgrade_timings["2400"]
 
 # DDR4 (SO-DIMM) -----------------------------------------------------------------------------------
-class KVR21SE15S84(SDRAMModule):
-    memtype = "DDR4"
+
+class KVR21SE15S84(DDR4Module):
     # geometry
     ngroupbanks = 4
     ngroups     = 4
@@ -776,8 +894,7 @@ class KVR21SE15S84(SDRAMModule):
     }
     speedgrade_timings["default"] = speedgrade_timings["2133"]
 
-class MTA4ATF51264HZ(SDRAMModule):
-    memtype = "DDR4"
+class MTA4ATF51264HZ(DDR4Module):
     # geometry
     ngroupbanks = 4
     ngroups     = 2
@@ -792,3 +909,21 @@ class MTA4ATF51264HZ(SDRAMModule):
         "2133": _SpeedgradeTimings(tRP=13.5, tRCD=13.5, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=33),
     }
     speedgrade_timings["default"] = speedgrade_timings["2133"]
+
+# DDR4 (RDIMM) -------------------------------------------------------------------------------------
+
+class MTA18ASF2G72PZ(DDR4RegisteredModule):
+    # geometry
+    ngroupbanks = 4
+    ngroups     = 4
+    nbanks      = ngroups * ngroupbanks
+    nrows       = 131072
+    ncols       = 1024
+    # timings
+    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
+    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
+    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 4.9), tZQCS=(128, 80))
+    speedgrade_timings = {
+        "2400": _SpeedgradeTimings(tRP=13.32, tRCD=13.32, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
+    }
+    speedgrade_timings["default"] = speedgrade_timings["2400"]
