@@ -497,8 +497,9 @@ class S7DDRPHY(Module, AutoCSR):
                 ]
 
             dq_bitslip = BitSlip(8,
-                rst = self._dly_sel.storage[i//8] & self._rdly_dq_bitslip_rst.re,
-                slp = self._dly_sel.storage[i//8] & self._rdly_dq_bitslip.re)
+                rst    = self._dly_sel.storage[i//8] & self._rdly_dq_bitslip_rst.re,
+                slp    = self._dly_sel.storage[i//8] & self._rdly_dq_bitslip.re,
+                cycles = 2)
             self.submodules += dq_bitslip
             self.comb += dq_bitslip.i.eq(dq_i_data)
             self.comb += [
